/*********************************************************************************
 *
 * File name:		defination_optical.h
 * Version:			4.0
 * Software:
 * Authors:			Zhehui Wang, Shixi Chen, Jiang Xu
 * Website:			https://eexu.home.ece.ust.hk/OEIL.html
 * The copyright information of this program can be found in the file COPYRIGHT.
 *
 *********************************************************************************/
#ifndef NONLINEARLOSSMODEL_H
#define NONLINEARLOSSMODEL_H

// #define __DEBUG_

#include <boost/array.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <boost/property_tree/ptree.hpp>
#include <cmath>
#include <iostream>
#include <vector>

#include <boost/numeric/odeint.hpp>
#include <fstream>

/*
 * References
 * [1] M. Dinu, F. Quochi, and H. Garcia, “Third-order nonlinearities in silicon at telecom wavelengths,” Appl. Phys. Lett., vol.
 * 82, no. 18, pp. 2954–2956, Apr. 2003. [2] K. Dietmar, Silicon-Organic Hybrid Platform for Photonic Integrated Circuits. KIT
 * Scientific Publishing, 2015.
 */

extern double laser_wavelength;
extern double propagation_loss;
extern double effective_mode_area;
extern double carrier_lifetime;
extern double TPA_coefficient;
extern double FCA_coefficient;

/* The type of container used to hold the state vector */
typedef std::array<double, 1> intensity_state;

namespace utils {
double loss_dB2scalar(double loss_in_dB);
double loss_scaler2dB(double loss_in_scalar);
}

struct push_back_intensity_and_distance {
    std::vector<intensity_state>& intensities;
    std::vector<double>& times;

    push_back_intensity_and_distance(std::vector<intensity_state>& t_intensities, std::vector<double>& t_times)
        : intensities(t_intensities)
        , times(t_times) {}

    void operator()(const intensity_state& intensity, double time) {
        intensities.push_back(intensity);
        times.push_back(time);
    }
};

class NonlinearLossModel {
  protected:
    /* Constant */
    const double kPhotonEnergy1550; // J
    const double kStartPosition;    // start from the head of a waveguide
    const double kWavelength;       // Default: C-band: 1.55um
    const double kStepLength;

    /* Input parameters */
    // major adjustables
    double alpha_dB_;            // Default: Linear loss: 0.2dB/cm
    double carrier_lifetime_;    // Default: 4e-9s
    double effective_mode_area_; // Default: 1e-8cm^2
    double stop_position_;       // Length of the wavelength in question. Default: 1.5cm
    double initial_power_;       // The coupled power at distance 0. Default: 100e-3W (100mW)

    // minor adjustables
    double beta_;  // Default: TPA coefficient at 1550nm [1]
    double sigma_; // Default: FCA coefficient: 1.45*10^17 cm2 [2]

    // derived model parameters
    double photon_energy_; // Default: hv at 1.55um is 1.28 * 10^-19 J
    double alpha_;
    double gamma_;
    double initial_intensity_;

    // result buffer
    double totalLoss_;

    void initParameters();
    void updateModel();

  public:
#ifdef __DEBUG_
    NonlinearLossModel();
#endif
    NonlinearLossModel(double coupled_in_power, double waveguideLength);

    virtual void echoParameters();

    // getters
    double getLinearLossPerLength() const { return alpha_dB_; }
    double getCarrierLifetime() const { return carrier_lifetime_; }
    double getEffectiveModeArea() const { return effective_mode_area_; }
    double getStartPosition() const { return kStartPosition; }
    double getStopPosition() const { return stop_position_; }
    double getInitialPower() const { return initial_power_; }
    double getInitialIntensity() const { return initial_intensity_; }

    // setters
    void setLinearLoss(double linearLoss_dB);
    void setCarrierLifetime(double lifetime);
    void setEffectiveModeArea(double Aeff);
    void setWaveguideLength(double length);
    void setInitialPower(double initalPower);

    // loaders
    void resetParameters();
    void setParametersFromFile();

    virtual void operator()(const intensity_state& I, intensity_state& dIdz, const double z) const;

    double totalLossAcrossWaveguide();
    double totalLossAcrossWaveguide(double initalPower);
    double linearLossAcrossWaveguide();
    double nonlinearLossAcrossWaveguide();
    double nonlinearLossAcrossWaveguide(double initalPower);
};

#endif